@@ -1,5 +1,5 @@
-/* AlsaMixer - Commandline mixer for the ALSA project
- * Copyright (C) 1998, 1999 Tim Janik <timj@gtk.org> and Jaroslav Kysela <perex@suse.cz>
+/* AlsaMixer - Commandline mixer for the ALSA project Copyright (C) 1998,
+ * 1999 Tim Janik <timj@gtk.org> and Jaroslav Kysela <perex@suse.cz>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -18,6 +18,15 @@
  *
  * ChangeLog:
  *
+ * Wed Feb 14 13:08:17 CET 2001   Jaroslav Kysela <perex@suse.cz>
+ *
+ *      * ported to the latest mixer 0.9.x API (function based)
+ *
+ * Fri Jun 23 14:10:00 MEST 2000  Jaroslav Kysela <perex@suse.cz>
+ *
+ *      * ported to new mixer 0.9.x API (simple control)
+ *      * improved error handling (mixer_abort)
+ *
  * Thu Mar  9 22:54:16 MET 2000  Takashi iwai <iwai@ww.uni-erlangen.de>
  *
  *	* a group is split into front, rear, center and woofer elements.
@@ -26,7 +35,7 @@
  *
  *      * version 1.00
  *
- *      * ported to new mixer API (group control)
+ *      * ported to new mixer API (scontrol control)
  *
  * Sun Feb 21 19:55:01 1999  Tim Janik  <timj@gtk.org>
  *
@@ -102,6 +111,7 @@
 #include <stdlib.h>
 #include <unistd.h>
 #include <sys/signal.h>
+#include <sys/time.h>
 
 #ifndef CURSESINC
 #include <ncurses.h>
@@ -110,7 +120,7 @@
 #endif
 #include <time.h>
 
-#include <sys/asoundlib.h>
+#include <alsa/asoundlib.h>
 
 /* example compilation commandline:
  * clear; gcc -Wall -pipe -O2 alsamixer.c -o alsamixer -lasound -lncurses
@@ -120,7 +130,7 @@
 #define	PRGNAME		 "alsamixer"
 #define	PRGNAME_UPPER	 "AlsaMixer"
 #define	VERSION		 "v1.00"
-#define	CHECK_ABORT(e,s) ({ if (errno != EINTR) mixer_abort ((e), (s)); })
+#define	CHECK_ABORT(e,s,n) ({ if ((n) != -EINTR) mixer_abort ((e), (s), (n)); })
 #define GETCH_BLOCK(w)	 ({ timeout ((w) ? -1 : 0); })
 
 #undef MAX
@@ -172,11 +182,10 @@ static int	 mixer_ofs_x = 0;
 static float	 mixer_extra_space = 0;
 static int	 mixer_cbar_height = 0;
 
-static int	 card_id = 0;
-static int	 mixer_id = 0;
+static char	 card_id[64] = "default";
 static snd_mixer_t *mixer_handle;
-static char	*mixer_card_name = NULL;
-static char	*mixer_device_name = NULL;
+static char	 mixer_card_name[128];
+static char	 mixer_device_name[128];
 
 /* mixer bar channel : left or right */
 #define MIXER_CHN_LEFT		0
@@ -190,29 +199,31 @@ static char	*mixer_device_name = NULL;
 enum {
   MIXER_ELEM_FRONT, MIXER_ELEM_REAR,
   MIXER_ELEM_CENTER, MIXER_ELEM_WOOFER,
+  MIXER_ELEM_CAPTURE,
+  MIXER_ELEM_ENUM,
   MIXER_ELEM_END
 };
 
-/* channel mask for each type */
-static int mixer_elem_mask[] = {
-  (SND_MIXER_CHN_MASK_FRONT_LEFT | SND_MIXER_CHN_MASK_FRONT_RIGHT),
-  (SND_MIXER_CHN_MASK_REAR_LEFT | SND_MIXER_CHN_MASK_REAR_RIGHT),
-  SND_MIXER_CHN_MASK_FRONT_CENTER,
-  SND_MIXER_CHN_MASK_WOOFER,
-};
+#define MIXER_ELEM_TYPE_MASK		0xff
+#define MIXER_ELEM_CAPTURE_SWITCH	0x100 /* bit */
+#define MIXER_ELEM_MUTE_SWITCH		0x200 /* bit */
+#define MIXER_ELEM_CAPTURE_SUFFIX	0x400
+#define MIXER_ELEM_HAS_VOLUME		0x800
 
 /* left and right channels for each type */
-static int mixer_elem_chn[][2] = {
-  { SND_MIXER_CHN_FRONT_LEFT, SND_MIXER_CHN_FRONT_RIGHT },
-  { SND_MIXER_CHN_REAR_LEFT, SND_MIXER_CHN_REAR_RIGHT },
-  { SND_MIXER_CHN_FRONT_CENTER, -1 },
-  { SND_MIXER_CHN_WOOFER, -1 },
+static snd_mixer_selem_channel_id_t mixer_elem_chn[][2] = {
+  { SND_MIXER_SCHN_FRONT_LEFT, SND_MIXER_SCHN_FRONT_RIGHT },
+  { SND_MIXER_SCHN_REAR_LEFT, SND_MIXER_SCHN_REAR_RIGHT },
+  { SND_MIXER_SCHN_FRONT_CENTER, SND_MIXER_SCHN_UNKNOWN },
+  { SND_MIXER_SCHN_WOOFER, SND_MIXER_SCHN_UNKNOWN },
+  { SND_MIXER_SCHN_FRONT_LEFT, SND_MIXER_SCHN_FRONT_RIGHT },
 };
 
-static snd_mixer_gid_t *mixer_gid = NULL;
-static int	 mixer_n_groups = 0;
+static void 	*mixer_sid = NULL;
+static int	 mixer_n_selems = 0;
+static int       mixer_changed_state = 1;
 
-/* split groups */
+/* split scontrols */
 static int	 mixer_n_elems = 0;
 static int	 mixer_n_vis_elems = 0;
 static int	 mixer_first_vis_elem = 0;
@@ -348,7 +359,8 @@ typedef enum
 /* --- prototypes --- */
 static void
 mixer_abort (ErrType error,
-	     const char *err_string)
+	     const char *err_string,
+	     int xerrno)
      __attribute__
 ((noreturn));
 
@@ -375,7 +387,8 @@ mixer_clear (int full_redraw)
 
 static void
 mixer_abort (ErrType     error,
-	     const char *err_string)
+	     const char *err_string,
+	     int	 xerrno)
 {
   if (mixer_window)
     {
@@ -392,16 +405,16 @@ mixer_abort (ErrType     error,
     {
     case ERR_OPEN:
       fprintf (stderr,
-	       PRGNAME ": failed to open mixer #%i/#%i: %s\n",
+	       PRGNAME ": function %s failed for %s: %s\n",
+	       err_string,
 	       card_id,
-	       mixer_id,
-	       snd_strerror (errno));
+	       snd_strerror (xerrno));
       break;
     case ERR_FCN:
       fprintf (stderr,
 	       PRGNAME ": function %s failed: %s\n",
 	       err_string,
-	       snd_strerror (errno));
+	       snd_strerror (xerrno));
       break;
     case ERR_SIGNAL:
       fprintf (stderr,
@@ -463,23 +476,58 @@ mixer_conv(int val, int omin, int omax, int nmin, int nmax)
 }
 
 static int
-mixer_calc_volume(snd_mixer_group_t *group, int vol, int chn)
+mixer_calc_volume(snd_mixer_elem_t *elem,
+		  int vol, int type,
+		  snd_mixer_selem_channel_id_t chn)
 {
   int vol1;
-
+  long v;
+  long min, max;
+  if (type != MIXER_ELEM_CAPTURE)
+    snd_mixer_selem_get_playback_volume_range(elem, &min, &max);
+  else
+    snd_mixer_selem_get_capture_volume_range(elem, &min, &max);
   vol1 = (vol < 0) ? -vol : vol;
   if (vol1 > 0) {
     if (vol1 > 100)
-      vol1 = group->max;
+      vol1 = max;
     else
-      vol1 = mixer_conv(vol1, 0, 100, group->min, group->max);
+      vol1 = mixer_conv(vol1, 0, 100, min, max);
+    /* Note: we have delta in vol1 and we need to map our */
+    /*       delta value to hardware range */
+    vol1 -= min;
     if (vol1 <= 0)
       vol1 = 1;
     if (vol < 0)
       vol1 = -vol1;
   }
-  vol1 += group->volume.values[chn];
-  return CLAMP(vol1, group->min, group->max);
+  if (type != MIXER_ELEM_CAPTURE)
+    snd_mixer_selem_get_playback_volume(elem, chn, &v);
+  else
+    snd_mixer_selem_get_capture_volume(elem, chn, &v);
+  vol1 += v;
+  return CLAMP(vol1, min, max);
+}
+
+/* update enum list */
+static void update_enum_list(snd_mixer_elem_t *elem, int chn, int delta)
+{
+  int eidx;
+  if (snd_mixer_selem_get_enum_item(elem, chn, &eidx) < 0)
+    return;
+  if (delta < 0) {
+    eidx--;
+    if (eidx < 0)
+      return;
+  } else {
+    int items = snd_mixer_selem_get_enum_items(elem);
+    if (items < 0)
+      return;
+    eidx++;
+    if (eidx >= items)
+      return;
+  }
+  snd_mixer_selem_set_enum_item(elem, chn, eidx);
 }
 
 /* set new channel values
@@ -487,141 +535,262 @@ mixer_calc_volume(snd_mixer_group_t *group, int vol, int chn)
 static void
 mixer_write_cbar (int elem_index)
 {
-  snd_mixer_group_t group;
+  snd_mixer_elem_t *elem;
   int vleft, vright, vbalance;
-  int type, chn_left, chn_right;
-  int i, changed;
+  int type;
+  snd_mixer_selem_id_t *sid;
+  snd_mixer_selem_channel_id_t chn_left, chn_right, chn;
+  int sw;
 
-  bzero(&group, sizeof(group));
-  group.gid = mixer_gid[mixer_grpidx[elem_index]];
-  if (snd_mixer_group_read (mixer_handle, &group) < 0)
-    CHECK_ABORT (ERR_FCN, "snd_mixer_group_read()");
+  if (mixer_sid == NULL)
+    return;
   
-  type = mixer_type[elem_index];
+  sid = (snd_mixer_selem_id_t *)(((char *)mixer_sid) + snd_mixer_selem_id_sizeof() * mixer_grpidx[elem_index]);
+  elem = snd_mixer_find_selem(mixer_handle, sid);
+  if (elem == NULL)
+    CHECK_ABORT (ERR_FCN, "snd_mixer_find_selem()", -EINVAL);
+  type = mixer_type[elem_index] & MIXER_ELEM_TYPE_MASK;
   chn_left = mixer_elem_chn[type][MIXER_CHN_LEFT];
-  if (! (group.channels & (1 << chn_left)))
-    return; /* ..??.. */
   chn_right = mixer_elem_chn[type][MIXER_CHN_RIGHT];
-  if (chn_right >= 0 && ! (group.channels & (1 << chn_right)))
-    chn_right = -1;
+  if (chn_right != SND_MIXER_SCHN_UNKNOWN) {
+    if (type != MIXER_ELEM_CAPTURE) {
+      if (!snd_mixer_selem_has_playback_channel(elem, chn_right))
+	chn_right = SND_MIXER_SCHN_UNKNOWN;
+    } else {
+      if (!snd_mixer_selem_has_capture_channel(elem, chn_right))
+	chn_right = SND_MIXER_SCHN_UNKNOWN;
+    }
+  }
 
-  changed = 0;
-
-  /* volue
+  /* volume
    */
   if ((mixer_volume_delta[MIXER_CHN_LEFT] ||
        mixer_volume_delta[MIXER_CHN_RIGHT] ||
        mixer_balance_volumes) &&
-      (group.caps & SND_MIXER_GRPCAP_VOLUME)) {
-    int mono = 
-      (chn_right < 0 || (group.caps & SND_MIXER_GRPCAP_JOINTLY_VOLUME));
+      (mixer_type[elem_index] & MIXER_ELEM_HAS_VOLUME)) {
+    int mono;
+    int joined;
+    mono = (chn_right == SND_MIXER_SCHN_UNKNOWN);
+    if (type != MIXER_ELEM_CAPTURE)
+      joined = snd_mixer_selem_has_playback_volume_joined(elem);
+    else
+	joined = snd_mixer_selem_has_capture_volume_joined(elem);
+    mono |= joined;
     if (mono && !mixer_volume_delta[MIXER_CHN_LEFT])
       mixer_volume_delta[MIXER_CHN_LEFT] = mixer_volume_delta[MIXER_CHN_RIGHT];
-    vleft = mixer_calc_volume(&group, mixer_volume_delta[MIXER_CHN_LEFT], chn_left);
+    vleft = mixer_calc_volume(elem, mixer_volume_delta[MIXER_CHN_LEFT], type, chn_left);
     vbalance = vleft;
     if (! mono) {
-      vright = mixer_calc_volume(&group, mixer_volume_delta[MIXER_CHN_RIGHT], chn_right);
+      vright = mixer_calc_volume(elem, mixer_volume_delta[MIXER_CHN_RIGHT], type, chn_right);
       vbalance += vright;
       vbalance /= 2;
-    } else
+    } else {
       vright = vleft;
-    if (vleft >= 0 && vright >= 0) {
-      if (group.caps & SND_MIXER_GRPCAP_JOINTLY_VOLUME) {
-	for (i = 0; i < SND_MIXER_CHN_LAST; i++) {
-	  if (group.channels & (1 << i))
-	    group.volume.values[i] = vleft;
-	}
+    }
+
+      if (joined) {
+	for (chn = 0; chn < SND_MIXER_SCHN_LAST; chn++)
+	  if (type != MIXER_ELEM_CAPTURE) {
+	    if (snd_mixer_selem_has_playback_channel(elem, chn))
+	      snd_mixer_selem_set_playback_volume(elem, chn, vleft);
+	  } else {
+	    if (snd_mixer_selem_has_capture_channel(elem, chn))
+	      snd_mixer_selem_set_capture_volume(elem, chn, vleft);
+	  }
       } else {
 	if (mixer_balance_volumes)
 	  vleft = vright = vbalance;
-	group.volume.values[chn_left] = vleft;
-	if (! mono)
-	  group.volume.values[chn_right] = vright;
+	if (type != MIXER_ELEM_CAPTURE) {
+	  if (snd_mixer_selem_has_playback_volume(elem) &&
+	      snd_mixer_selem_has_playback_channel(elem, chn_left))
+	    snd_mixer_selem_set_playback_volume(elem, chn_left, vleft);
+	} else {
+	  if (snd_mixer_selem_has_capture_volume(elem) &&
+	      snd_mixer_selem_has_capture_channel(elem, chn_left))
+	    snd_mixer_selem_set_capture_volume(elem, chn_left, vleft);
+	}
+	if (! mono) {
+	  if (type != MIXER_ELEM_CAPTURE) {
+	    if (snd_mixer_selem_has_playback_volume(elem) &&
+		snd_mixer_selem_has_playback_channel(elem, chn_right))
+	      snd_mixer_selem_set_playback_volume(elem, chn_right, vright);
+	  } else {
+	    if (snd_mixer_selem_has_capture_volume(elem) &&
+		snd_mixer_selem_has_capture_channel(elem, chn_right))
+	      snd_mixer_selem_set_capture_volume(elem, chn_right, vright);
+	  }
+	}
       }
-      changed = 1;
-    }
   }
-  mixer_volume_delta[MIXER_CHN_LEFT] = mixer_volume_delta[MIXER_CHN_RIGHT] = 0;
-  mixer_balance_volumes = 0;
 
   /* mute
    */
-  if (mixer_toggle_mute && (group.caps & SND_MIXER_GRPCAP_MUTE)) {
-    group.mute &= group.channels;
-    if (group.caps & SND_MIXER_GRPCAP_JOINTLY_MUTE)
-      group.mute = group.mute ? 0 : group.channels;
-    else {
-      if (mixer_toggle_mute & MIXER_MASK_LEFT)
-	group.mute ^= (1 << chn_left);
-      if (chn_right >= 0 && (mixer_toggle_mute & MIXER_MASK_RIGHT))
-	group.mute ^= (1 << chn_right);
+  if (mixer_type[elem_index] & MIXER_ELEM_MUTE_SWITCH) {
+    if (mixer_toggle_mute && snd_mixer_selem_has_playback_switch(elem)) {
+      if (snd_mixer_selem_has_playback_switch_joined(elem)) {
+	snd_mixer_selem_get_playback_switch(elem, chn_left, &sw);
+	snd_mixer_selem_set_playback_switch_all(elem, !sw);
+      } else {
+	if (mixer_toggle_mute & MIXER_MASK_LEFT) {
+	  snd_mixer_selem_get_playback_switch(elem, chn_left, &sw);
+	  snd_mixer_selem_set_playback_switch(elem, chn_left, !sw);
+	}
+	if (chn_right != SND_MIXER_SCHN_UNKNOWN && 
+	    (mixer_toggle_mute & MIXER_MASK_RIGHT)) {
+	  snd_mixer_selem_get_playback_switch(elem, chn_right, &sw);
+	  snd_mixer_selem_set_playback_switch(elem, chn_right, !sw);
+	}
+      }
     }
-    changed = 1;
   }
   mixer_toggle_mute = 0;
 
   /* capture
    */
-  if (mixer_toggle_capture && (group.caps & SND_MIXER_GRPCAP_CAPTURE)) {
-    group.capture &= group.channels;
-    if (group.caps & SND_MIXER_GRPCAP_JOINTLY_CAPTURE)
-      group.capture = group.capture ? 0 : group.channels;
-    else {
-      if (mixer_toggle_capture & MIXER_MASK_LEFT)
-	group.capture ^= (1 << chn_left);
-      if (chn_right >= 0 && (mixer_toggle_capture & MIXER_MASK_RIGHT))
-	group.capture ^= (1 << chn_right);
+  if (mixer_type[elem_index] & MIXER_ELEM_CAPTURE_SWITCH) {
+    if (mixer_toggle_capture && snd_mixer_selem_has_capture_switch(elem)) {
+      if (snd_mixer_selem_has_capture_switch_joined(elem)) {
+	snd_mixer_selem_get_capture_switch(elem, chn_left, &sw);
+	snd_mixer_selem_set_capture_switch_all(elem, !sw);
+      } else {
+	if ((mixer_toggle_capture & MIXER_MASK_LEFT) &&
+	    snd_mixer_selem_has_capture_channel(elem, chn_left)) {
+	  snd_mixer_selem_get_capture_switch(elem, chn_left, &sw);
+	  snd_mixer_selem_set_capture_switch(elem, chn_left, !sw);
+	}
+	if (chn_right != SND_MIXER_SCHN_UNKNOWN && 
+	    snd_mixer_selem_has_capture_channel(elem, chn_right) &&
+	    (mixer_toggle_capture & MIXER_MASK_RIGHT)) {
+	  snd_mixer_selem_get_capture_switch(elem, chn_right, &sw);
+	  snd_mixer_selem_set_capture_switch(elem, chn_right, !sw);
+	}
+      }
     }
-    changed = 1;
   }
   mixer_toggle_capture = 0;
-      
-  if (changed) {
-    if (snd_mixer_group_write (mixer_handle, &group) < 0)
-      CHECK_ABORT (ERR_FCN, "snd_mixer_group_write()");
+
+  /* enum list
+   */
+  if (type == MIXER_ELEM_ENUM) {
+    if (mixer_volume_delta[MIXER_CHN_LEFT])
+      update_enum_list(elem, MIXER_CHN_LEFT, mixer_volume_delta[MIXER_CHN_LEFT]);
+    if (mixer_volume_delta[MIXER_CHN_RIGHT])
+      update_enum_list(elem, MIXER_CHN_RIGHT, mixer_volume_delta[MIXER_CHN_RIGHT]);
   }
+
+  mixer_volume_delta[MIXER_CHN_LEFT] = mixer_volume_delta[MIXER_CHN_RIGHT] = 0;
+  mixer_balance_volumes = 0;
 }
 
 
+static void display_enum_list(snd_mixer_elem_t *elem, int y, int x)
+{
+  int i, cury, ch, err;
+
+  /* clear */
+  mixer_dc(DC_TEXT);
+  for (i = mixer_cbar_height + 3, cury = y; i > 0; i--, cury--)
+    mvaddstr(cury, x, "       ");
+  
+  cury = y - 4;
+  for (ch = 0; ch < 2; ch++) {
+    int eidx, ofs;
+    char tmp[9];
+    err = snd_mixer_selem_get_enum_item(elem, ch, &eidx);
+    if (err < 0)
+      break;
+    if (snd_mixer_selem_get_enum_item_name(elem, eidx, sizeof(tmp) - 1, tmp) < 0)
+      break;
+    tmp[8] = 0;
+    ofs = (8 - strlen(tmp)) / 2;
+    mvaddstr(cury, x + ofs, tmp);
+    cury += 2;
+  }
+}
+
 static void
 mixer_update_cbar (int elem_index)
 {
-  char string[64], string1[64];
+  char string[128], string1[64], *suffix;
   int dc;
-  snd_mixer_group_t group;
-  int vleft, vright;
-  int type, chn_left, chn_right;
+  snd_mixer_elem_t *elem;
+  long vleft, vright;
+  int type;
+  snd_mixer_selem_id_t *sid;
+  snd_mixer_selem_channel_id_t chn_left, chn_right;
   int x, y, i;
+  int swl, swr;
 
-  /* set new group indices and read info
+  /* set new scontrol indices and read info
    */
-  bzero(&group, sizeof(group));
-  group.gid = mixer_gid[mixer_grpidx[elem_index]];
-  if (snd_mixer_group_read (mixer_handle, &group) < 0)
-    CHECK_ABORT (ERR_FCN, "snd_mixer_group_read()");
-  
-  type = mixer_type[elem_index];
+  if (mixer_sid == NULL)
+    return;
+
+  sid = (snd_mixer_selem_id_t *)(((char *)mixer_sid) + snd_mixer_selem_id_sizeof() * mixer_grpidx[elem_index]);
+  elem = snd_mixer_find_selem(mixer_handle, sid);
+  if (elem == NULL)
+    CHECK_ABORT (ERR_FCN, "snd_mixer_find_selem()", -EINVAL);
+
+  type = mixer_type[elem_index] & MIXER_ELEM_TYPE_MASK;
   chn_left = mixer_elem_chn[type][MIXER_CHN_LEFT];
-  if (! (group.channels & (1 << chn_left)))
-    return; /* ..??.. */
   chn_right = mixer_elem_chn[type][MIXER_CHN_RIGHT];
-  if (chn_right >= 0 && ! (group.channels & (1 << chn_right)))
-    chn_right = -1;
-  
-  /* first, read values for the numbers to be displayed
-   */
-  if (snd_mixer_group_read (mixer_handle, &group) < 0)
-    CHECK_ABORT (ERR_FCN, "snd_mixer_group_read()");
-  
-  vleft = group.volume.values[chn_left];
-  vleft = mixer_conv(vleft, group.min, group.max, 0, 100);
-  if (chn_right >= 0) {
-    vright = group.volume.values[chn_right];
-    vright = mixer_conv(vright, group.min, group.max, 0, 100);
-  } else {
-    vright = vleft;
+  if (chn_right != SND_MIXER_SCHN_UNKNOWN) {
+    if (type != MIXER_ELEM_CAPTURE) {
+      if (!snd_mixer_selem_has_playback_channel(elem, chn_right))
+	chn_right = SND_MIXER_SCHN_UNKNOWN;
+    } else {
+      if (!snd_mixer_selem_has_capture_channel(elem, chn_right))
+	chn_right = SND_MIXER_SCHN_UNKNOWN;
+    }
+  }
+
+  vleft = vright = 0;
+  if (type != MIXER_ELEM_CAPTURE && snd_mixer_selem_has_playback_volume(elem)) {
+    long vmin, vmax;
+    snd_mixer_selem_get_playback_volume_range(elem, &vmin, &vmax);
+    snd_mixer_selem_get_playback_volume(elem, chn_left, &vleft);
+    vleft = mixer_conv(vleft, vmin, vmax, 0, 100);
+    if (chn_right != SND_MIXER_SCHN_UNKNOWN) {
+      snd_mixer_selem_get_playback_volume(elem, chn_right, &vright);
+      vright = mixer_conv(vright, vmin, vmax, 0, 100);
+    } else {
+      vright = vleft;
+    }
   }
   
+  if (type == MIXER_ELEM_CAPTURE && snd_mixer_selem_has_capture_volume(elem)) {
+    long vmin, vmax;
+    snd_mixer_selem_get_capture_volume_range(elem, &vmin, &vmax);
+    snd_mixer_selem_get_capture_volume(elem, chn_left, &vleft);
+    vleft = mixer_conv(vleft, vmin, vmax, 0, 100);
+    if (chn_right != SND_MIXER_SCHN_UNKNOWN) {
+      snd_mixer_selem_get_capture_volume(elem, chn_right, &vright);
+      vright = mixer_conv(vright, vmin, vmax, 0, 100);
+    } else {
+      vright = vleft;
+    }
+  }
+  
+  /* update the focused full bar name
+   */
+  if (elem_index == mixer_focus_elem) {
+    mixer_dc (DC_PROMPT);
+    mvaddstr (3, 2, "Item: ");
+    mixer_dc (DC_TEXT);
+    string1[8] = 0;
+    for (i = 0; i < 63; i++)
+      string1[i] = ' ';
+    string1[63] = '\0';
+    strcpy(string, snd_mixer_selem_id_get_name(sid));
+    if (mixer_type[elem_index] & MIXER_ELEM_CAPTURE_SUFFIX)
+      strcat(string, " Capture");
+    if (snd_mixer_selem_id_get_index(sid) > 0)
+      sprintf(string + strlen(string), " %i", snd_mixer_selem_id_get_index(sid));
+    string[63] = '\0';
+    strncpy(string1, string, strlen(string));
+    addstr(string1);
+  }
+
   /* get channel bar position
    */
   if (!mixer_cbar_get_pos (elem_index, &x, &y))
@@ -630,10 +799,14 @@ mixer_update_cbar (int elem_index)
   /* channel bar name
    */
   mixer_dc (elem_index == mixer_focus_elem ? DC_CBAR_FOCUS_LABEL : DC_CBAR_LABEL);
-  if (group.gid.index > 0)
-    sprintf(string1, "%s %d", group.gid.name, group.gid.index);
+  if (mixer_type[elem_index] & MIXER_ELEM_CAPTURE_SUFFIX)
+    suffix = " Capture";
   else
-    strcpy(string1, group.gid.name);
+    suffix = "";
+  if (snd_mixer_selem_id_get_index(sid) > 0)
+    sprintf(string1, "%s%s %d", snd_mixer_selem_id_get_name(sid), suffix, snd_mixer_selem_id_get_index(sid));
+  else
+    sprintf(string1, "%s%s", snd_mixer_selem_id_get_name(sid), suffix);
   string1[8] = 0;
   for (i = 0; i < 8; i++)
     {
@@ -648,17 +821,24 @@ mixer_update_cbar (int elem_index)
    */
   mixer_dc (DC_BACK);
   mvaddstr (y, x, "         ");
-  mixer_dc (DC_TEXT);
-  sprintf (string, "%d", vleft);
-  mvaddstr (y, x + 3 - strlen (string), string);
-  mixer_dc (DC_CBAR_FRAME);
-  mvaddch (y, x + 3, '<');
-  mvaddch (y, x + 4, '>');
-  mixer_dc (DC_TEXT);
-  sprintf (string, "%d", vright);
-  mvaddstr (y, x + 5, string);
+  if (mixer_type[elem_index] & MIXER_ELEM_HAS_VOLUME) {
+    mixer_dc (DC_TEXT);
+    sprintf (string, "%ld", vleft);
+    mvaddstr (y, x + 3 - strlen (string), string);
+    mixer_dc (DC_CBAR_FRAME);
+    mvaddch (y, x + 3, '<');
+    mvaddch (y, x + 4, '>');
+    mixer_dc (DC_TEXT);
+    sprintf (string, "%ld", vright);
+    mvaddstr (y, x + 5, string);
+  }
   y--;
   
+  if (type == MIXER_ELEM_ENUM) {
+    display_enum_list(elem, y, x);
+    return;
+  } 
+
   /* left/right bar
    */
   mixer_dc (DC_CBAR_FRAME);
@@ -692,40 +872,56 @@ mixer_update_cbar (int elem_index)
    */
   mixer_dc (DC_BACK);
   mvaddstr (y, x, "         ");
-  mixer_dc (DC_CBAR_FRAME);
-  mvaddch (y, x + 2, ACS_ULCORNER);
-  dc = group.mute & (1 << chn_left) ? DC_CBAR_MUTE : DC_CBAR_NOMUTE;
-  mvaddch (y, x + 3, mixer_dc (dc));
-  if (chn_right >= 0)
-    dc = group.mute & (1 << chn_right) ? DC_CBAR_MUTE : DC_CBAR_NOMUTE;
-  mvaddch (y, x + 4, mixer_dc (dc));
-  mixer_dc (DC_CBAR_FRAME);
-  mvaddch (y, x + 5, ACS_URCORNER);
+  if ((mixer_type[elem_index] & MIXER_ELEM_MUTE_SWITCH)
+      && snd_mixer_selem_has_playback_switch(elem)) {
+    mixer_dc (DC_CBAR_FRAME);
+    mvaddch (y, x + 2, ACS_ULCORNER);
+    snd_mixer_selem_get_playback_switch(elem, chn_left, &swl);
+    dc = swl ? DC_CBAR_NOMUTE : DC_CBAR_MUTE;
+    mvaddch (y, x + 3, mixer_dc (dc));
+    if (chn_right != SND_MIXER_SCHN_UNKNOWN) {
+      snd_mixer_selem_get_playback_switch(elem, chn_right, &swr);
+      dc = swr ? DC_CBAR_NOMUTE : DC_CBAR_MUTE;
+    }
+    mvaddch (y, x + 4, mixer_dc (dc));
+    mixer_dc (DC_CBAR_FRAME);
+    mvaddch (y, x + 5, ACS_URCORNER);
+  } else {
+    mixer_dc (DC_CBAR_FRAME);
+    mvaddch (y, x + 2, ACS_ULCORNER);
+    mvaddch (y, x + 3, ACS_HLINE);
+    mvaddch (y, x + 4, ACS_HLINE);
+    mvaddch (y, x + 5, ACS_URCORNER);
+  }    
   y--;
   
   /* capture input?
    */
-  if ((group.capture & (1 << chn_left)) ||
-      (chn_right >= 0 && (group.capture & (1 << chn_right))))
-    {
+  if ((mixer_type[elem_index] & MIXER_ELEM_CAPTURE_SWITCH) &&
+      snd_mixer_selem_has_capture_switch(elem)) {
+    int has_r_sw = chn_right != SND_MIXER_SCHN_UNKNOWN &&
+      snd_mixer_selem_has_capture_channel(elem, chn_right);
+    snd_mixer_selem_get_capture_switch(elem, chn_left, &swl);
+    if (has_r_sw)
+      snd_mixer_selem_get_capture_switch(elem, chn_right, &swr);
+    if (swl || (has_r_sw && swr)) {
       mixer_dc (DC_CBAR_CAPTURE);
       mvaddstr (y, x + 1, "CAPTUR");
-      if (group.capture & (1 << chn_left)) {
+      if (swl) {
 	mvaddstr (y + 1, x + 1, "L");
-	if (chn_right < 0)
+	if (! has_r_sw)
 	  mvaddstr (y + 1, x + 6, "R");
       }
-      if (chn_right >= 0 && (group.capture & (1 << chn_right)))
+      if (has_r_sw && swr)
 	mvaddstr (y + 1, x + 6, "R");
+    } else {
+      for (i = 0; i < 6; i++)
+	mvaddch (y, x + 1 + i, mixer_dc (DC_CBAR_NOCAPTURE));
     }
-  else if (group.caps & SND_MIXER_GRPCAP_CAPTURE)
-    for (i = 0; i < 6; i++)
-      mvaddch (y, x + 1 + i, mixer_dc (DC_CBAR_NOCAPTURE));
-  else
-    {
-      mixer_dc (DC_BACK);
-      mvaddstr (y, x, "         ");
-    }
+  } else {
+    mixer_dc (DC_BACK);
+    mvaddstr (y, x, "         ");
+  }
   y--;
 }
 
@@ -736,6 +932,7 @@ mixer_update_cbars (void)
   static int o_y = 0;
   int i, x, y;
   
+  
   if (!mixer_cbar_get_pos (mixer_focus_elem, &x, &y))
     {
       if (mixer_focus_elem < mixer_first_vis_elem)
@@ -744,9 +941,18 @@ mixer_update_cbars (void)
 	mixer_first_vis_elem = mixer_focus_elem - mixer_n_vis_elems + 1;
       mixer_cbar_get_pos (mixer_focus_elem, &x, &y);
     }
+  if (mixer_first_vis_elem + mixer_n_vis_elems >= mixer_n_elems) {
+     mixer_first_vis_elem = mixer_n_elems - mixer_n_vis_elems;
+     if (mixer_first_vis_elem < 0)
+       mixer_first_vis_elem = 0;
+     mixer_cbar_get_pos (mixer_focus_elem, &x, &y);
+  }
   mixer_write_cbar(mixer_focus_elem);
-  for (i = 0; i < mixer_n_vis_elems; i++)
+  for (i = 0; i < mixer_n_vis_elems; i++) {
+    if (i + mixer_first_vis_elem >= mixer_n_elems)
+      continue;
     mixer_update_cbar (i + mixer_first_vis_elem);
+  }
   
   /* draw focused cbar
    */
@@ -780,7 +986,7 @@ mixer_draw_frame (void)
   mixer_dc (DC_TEXT);
   sprintf (string, "%s", mixer_card_name);
   max_len = mixer_max_x - 2 - 6 - 2;
-  if (strlen (string) > max_len)
+  if ((int)strlen (string) > max_len)
     string[max_len] = 0;
   addstr (string);
   
@@ -791,7 +997,7 @@ mixer_draw_frame (void)
   mixer_dc (DC_TEXT);
   sprintf (string, "%s", mixer_device_name);
   max_len = mixer_max_x - 2 - 6 - 2;
-  if (strlen (string) > max_len)
+  if ((int)strlen (string) > max_len)
     string[max_len] = 0;
   addstr (string);
 
@@ -826,7 +1032,7 @@ mixer_draw_frame (void)
 
   /* program title
    */
-  sprintf (string, "%s %s", PRGNAME_UPPER, VERSION);
+  sprintf (string, "%s %s (Press Escape to quit)", PRGNAME_UPPER, VERSION);
   max_len = strlen (string);
   if (mixer_max_x >= max_len + 4)
     {
@@ -885,7 +1091,8 @@ mixer_show_text (char *title,
   float hoffs, voffs;
   char *p, *text_offs = text;
   int x1, x2, y1, y2;
-  int i, n, l, r, block, stipple;
+  int i, n, l, r;
+  unsigned long block, stipple;
 
   /* coords
    */
@@ -1174,121 +1381,187 @@ mixer_show_procinfo (void)
   vbuffer_kill (&vbuf);
 }
 
+static int
+mixer_event (snd_mixer_t *mixer, unsigned int mask, snd_mixer_elem_t *elem)
+{
+  mixer_changed_state = 1;
+  return 0;
+}
+
 static void
 mixer_init (void)
 {
-  static snd_mixer_info_t mixer_info = { 0, };
-  static struct snd_ctl_hw_info hw_info;
+  snd_ctl_card_info_t *hw_info;
   snd_ctl_t *ctl_handle;
+  int err;
+  snd_ctl_card_info_alloca(&hw_info);
   
-  if (snd_ctl_open (&ctl_handle, card_id) < 0)
-    mixer_abort (ERR_OPEN, "snd_ctl_open");
-  if (snd_ctl_hw_info (ctl_handle, &hw_info) < 0)
-    mixer_abort (ERR_FCN, "snd_ctl_hw_info");
+  if ((err = snd_ctl_open (&ctl_handle, card_id, 0)) < 0)
+    mixer_abort (ERR_OPEN, "snd_ctl_open", err);
+  if ((err = snd_ctl_card_info (ctl_handle, hw_info)) < 0)
+    mixer_abort (ERR_FCN, "snd_ctl_card_info", err);
   snd_ctl_close (ctl_handle);
   /* open mixer device
    */
-  if (snd_mixer_open (&mixer_handle, card_id, mixer_id) < 0)
-    mixer_abort (ERR_OPEN, "snd_mixer_open");
+  if ((err = snd_mixer_open (&mixer_handle, 0)) < 0)
+    mixer_abort (ERR_FCN, "snd_mixer_open", err);
+  if ((err = snd_mixer_attach (mixer_handle, card_id)) < 0)
+    mixer_abort (ERR_FCN, "snd_mixer_attach", err);
+  if ((err = snd_mixer_selem_register (mixer_handle, NULL, NULL)) < 0)
+    mixer_abort (ERR_FCN, "snd_mixer_selem_register", err);
+  snd_mixer_set_callback (mixer_handle, mixer_event);
+  if ((err = snd_mixer_load (mixer_handle)) < 0)
+    mixer_abort (ERR_FCN, "snd_mixer_load", err);
   
   /* setup global variables
    */
-  if (snd_mixer_info (mixer_handle, &mixer_info) < 0)
-    mixer_abort (ERR_FCN, "snd_mixer_info");
-  mixer_card_name = hw_info.name;
-  mixer_device_name = mixer_info.name;
+  strcpy(mixer_card_name, snd_ctl_card_info_get_name(hw_info));
+  strcpy(mixer_device_name, snd_ctl_card_info_get_mixername(hw_info));
 }
 
 static void
 mixer_reinit (void)
 {
-  snd_mixer_groups_t groups;
-  int idx, elem_index, i;
-  snd_mixer_gid_t focus_gid;
+  snd_mixer_elem_t *elem;
+  int idx, elem_index, i, j, selem_count;
+  snd_mixer_selem_id_t *sid;
+  snd_mixer_selem_id_t *focus_gid;
   int focus_type = -1;
+  snd_mixer_selem_id_alloca(&focus_gid);
   
-  if (mixer_gid) {
-    focus_gid = mixer_gid[mixer_grpidx[mixer_focus_elem]];
-    focus_type = mixer_type[mixer_focus_elem];
+  if (!mixer_changed_state)
+    return;
+  if (mixer_sid) {
+    snd_mixer_selem_id_copy(focus_gid, (snd_mixer_selem_id_t *)(((char *)mixer_sid) + snd_mixer_selem_id_sizeof() * mixer_grpidx[mixer_focus_elem]));
+    focus_type = mixer_type[mixer_focus_elem] & MIXER_ELEM_TYPE_MASK;
   }
-  while (1) {
-    bzero(&groups, sizeof(groups));
-    if (snd_mixer_groups(mixer_handle, &groups) < 0)
-      mixer_abort (ERR_FCN, "snd_mixer_groups");
-    mixer_n_groups = groups.groups_over;
-    if (mixer_n_groups > 0) {
-      groups.groups_size = mixer_n_groups;
-      groups.pgroups = (snd_mixer_gid_t *)malloc(sizeof(snd_mixer_gid_t) * mixer_n_groups);
-      if (groups.pgroups == NULL)
-        mixer_abort (ERR_FCN, "malloc");
-      groups.groups_over = 0;
-      groups.groups = 0;
-      if (snd_mixer_groups(mixer_handle, &groups) < 0)
-        mixer_abort (ERR_FCN, "snd_mixer_groups");
-      if (groups.groups_over > 0) {
-        free(groups.pgroups);
-        continue;
-      }
-    }
-    if (mixer_gid)
-      free(mixer_gid);
-    mixer_gid = groups.pgroups;
-    break;
+__again:
+  mixer_changed_state = 0;
+  if (mixer_sid != NULL)
+    free(mixer_sid);
+  selem_count = snd_mixer_get_count(mixer_handle);
+  mixer_sid = malloc(snd_mixer_selem_id_sizeof() * selem_count);
+  if (mixer_sid == NULL)
+    mixer_abort (ERR_FCN, "malloc", 0);
+  
+  mixer_n_selems = 0;
+  for (elem = snd_mixer_first_elem(mixer_handle); elem; elem = snd_mixer_elem_next(elem)) {
+    sid = (snd_mixer_selem_id_t *)(((char *)mixer_sid) + snd_mixer_selem_id_sizeof() * mixer_n_selems);
+    if (mixer_changed_state)
+      goto __again;
+    if (!snd_mixer_selem_is_active(elem))
+      continue;
+    snd_mixer_selem_get_id(elem, sid);
+    mixer_n_selems++;
   }
-  snd_mixer_sort_gid_table(mixer_gid, mixer_n_groups, snd_mixer_default_weights);
 
   mixer_n_elems = 0;
-  for (idx = 0; idx < mixer_n_groups; idx++) {
-    snd_mixer_group_t group;
-    bzero(&group, sizeof(group));
-    group.gid = mixer_gid[idx];
-    if (snd_mixer_group_read(mixer_handle, &group) < 0)
-      CHECK_ABORT (ERR_FCN, "snd_mixer_group_read()");
-    for (i = 0; i < MIXER_ELEM_END; i++) {
-      if (group.channels & mixer_elem_mask[i])
+  for (idx = 0; idx < mixer_n_selems; idx++) {
+    int nelems_added = 0;
+    sid = (snd_mixer_selem_id_t *)(((char *)mixer_sid) + snd_mixer_selem_id_sizeof() * idx);
+    if (mixer_changed_state)
+      goto __again;
+    elem = snd_mixer_find_selem(mixer_handle, sid);
+    if (elem == NULL)
+      CHECK_ABORT (ERR_FCN, "snd_mixer_find_selem()", -EINVAL);
+    for (i = 0; i < MIXER_ELEM_CAPTURE; i++) {
+      int ok;
+      for (j = ok = 0; j < 2; j++) {
+	if (mixer_changed_state)
+	  goto __again;
+	if (snd_mixer_selem_has_playback_channel(elem, mixer_elem_chn[i][j]))
+	  ok++;
+      }
+      if (ok) {
+	nelems_added++;
 	mixer_n_elems++;
+      }
     }
+    if (snd_mixer_selem_has_capture_volume(elem) ||
+	(nelems_added == 0 && snd_mixer_selem_has_capture_switch(elem)))
+      mixer_n_elems++;
   }
 
   if (mixer_type)
     free(mixer_type);
   mixer_type = (int *)malloc(sizeof(int) * mixer_n_elems);
   if (mixer_type == NULL)
-    mixer_abort(ERR_FCN, "malloc");
+    mixer_abort(ERR_FCN, "malloc", 0);
   if (mixer_grpidx)
     free(mixer_grpidx);
   mixer_grpidx = (int *)malloc(sizeof(int) * mixer_n_elems);
   if (mixer_grpidx == NULL)
-    mixer_abort(ERR_FCN, "malloc");
+    mixer_abort(ERR_FCN, "malloc", 0);
   elem_index = 0;
-  for (idx = 0; idx < mixer_n_groups; idx++) {
-    snd_mixer_group_t group;
-    bzero(&group, sizeof(group));
-    group.gid = mixer_gid[idx];
-    if (snd_mixer_group_read(mixer_handle, &group) < 0)
-      CHECK_ABORT (ERR_FCN, "snd_mixer_group_read()");
-    for (i = 0; i < MIXER_ELEM_END; i++) {
-      if (group.channels & mixer_elem_mask[i]) {
+  for (idx = 0; idx < mixer_n_selems; idx++) {
+    int nelems_added = 0;
+    sid = (snd_mixer_selem_id_t *)(((char *)mixer_sid) + snd_mixer_selem_id_sizeof() * idx);
+    if (mixer_changed_state)
+      goto __again;
+    elem = snd_mixer_find_selem(mixer_handle, sid);
+    if (elem == NULL)
+      CHECK_ABORT (ERR_FCN, "snd_mixer_find_selem()", -EINVAL);
+    for (i = 0; i < MIXER_ELEM_CAPTURE; i++) {
+      int ok;
+      for (j = ok = 0; j < 2; j++) {
+	if (mixer_changed_state)
+	  goto __again;
+	if (snd_mixer_selem_has_playback_channel(elem, mixer_elem_chn[i][j]))
+	  ok++;
+      }
+      if (ok) {
+	sid = (snd_mixer_selem_id_t *)(((char *)mixer_sid) + snd_mixer_selem_id_sizeof() * idx);
 	mixer_grpidx[elem_index] = idx;
-	mixer_type[elem_index] = i;
+	if (snd_mixer_selem_is_enumerated(elem)) {
+	  mixer_type[elem_index] = MIXER_ELEM_ENUM;
+	} else {
+	  mixer_type[elem_index] = i;
+	  if (i == 0 && snd_mixer_selem_has_playback_switch(elem))
+	    mixer_type[elem_index] |= MIXER_ELEM_MUTE_SWITCH;
+	  if (snd_mixer_selem_has_playback_volume(elem))
+	    mixer_type[elem_index] |= MIXER_ELEM_HAS_VOLUME;
+	  if (i == 0 && snd_mixer_selem_has_capture_switch(elem))
+	    mixer_type[elem_index] |= MIXER_ELEM_CAPTURE_SWITCH;
+	}
 	elem_index++;
+	nelems_added++;
 	if (elem_index >= mixer_n_elems)
 	  break;
       }
     }
+    if (snd_mixer_selem_has_capture_volume(elem) ||
+	(nelems_added == 0 && snd_mixer_selem_has_capture_switch(elem))) {
+      mixer_grpidx[elem_index] = idx;
+      mixer_type[elem_index] = MIXER_ELEM_CAPTURE;
+      if (nelems_added == 0 && snd_mixer_selem_has_capture_switch(elem))
+	mixer_type[elem_index] |= MIXER_ELEM_CAPTURE_SWITCH;
+      if (nelems_added)
+	mixer_type[elem_index] |= MIXER_ELEM_CAPTURE_SUFFIX;
+      if (snd_mixer_selem_has_capture_volume(elem))
+	mixer_type[elem_index] |= MIXER_ELEM_HAS_VOLUME;
+      elem_index++;
+      if (elem_index >= mixer_n_elems)
+	break;
+    }
   }
 
   mixer_focus_elem = 0;
   if (focus_type >= 0) {
     for (elem_index = 0; elem_index < mixer_n_elems; elem_index++) {
-      if (!memcmp(&focus_gid, &mixer_gid[mixer_grpidx[elem_index]], sizeof(focus_gid)) &&
-	  mixer_type[elem_index] == focus_type) {
+      sid = (snd_mixer_selem_id_t *)(((char *)mixer_sid) + snd_mixer_selem_id_sizeof() * mixer_grpidx[elem_index]);
+      if (!strcmp(snd_mixer_selem_id_get_name(focus_gid),
+                  snd_mixer_selem_id_get_name(sid)) &&
+          snd_mixer_selem_id_get_index(focus_gid) ==
+          snd_mixer_selem_id_get_index(sid) &&
+	  (mixer_type[elem_index] & MIXER_ELEM_TYPE_MASK) == focus_type) {
         mixer_focus_elem = elem_index;
         break;
       }
     }
   }
 
+  if (mixer_changed_state)
+    goto __again;
 }
 
 static void
@@ -1297,6 +1570,7 @@ mixer_init_window (void)
   /* initialize ncurses
    */
   mixer_window = initscr ();
+  curs_set (0); /* hide the cursor */
 
   mixer_no_lrcorner = tigetflag ("xenl") != 1 && tigetflag ("am") != 1;
 
@@ -1369,21 +1643,6 @@ mixer_resize (void)
     }
 }
 
-static void
-mixer_callback_rebuild (void *private_data)
-{
-  /* we don't actually need to update the individual channels because
-   * we redraw the whole screen upon every main iteration anyways.
-   */
-  mixer_reinit ();
-}
-
-static void
-mixer_callback_group (void *private_data, int cmd, snd_mixer_gid_t *gid)
-{
-  mixer_reinit ();
-}
-
 static void
 mixer_set_delta(int delta)
 {
@@ -1405,45 +1664,60 @@ mixer_add_delta(int delta)
 static int
 mixer_iteration (void)
 {
-  struct timeval delay = { 0, };
-  snd_mixer_callbacks_t callbacks = { 0, };
-  int mixer_fd;
-  fd_set rfds;
+  int count, err;
+  struct pollfd *fds;
   int finished = 0;
   int key = 0;
   int old_view;
+  unsigned short revents;
   
-  callbacks.rebuild = mixer_callback_rebuild;
-  callbacks.group = mixer_callback_group;
-
   /* setup for select on stdin and the mixer fd */
-  mixer_fd = snd_mixer_file_descriptor (mixer_handle);
-  FD_ZERO (&rfds);
-  FD_SET (fileno (stdin), &rfds);
-  FD_SET (mixer_fd, &rfds);
-
-  delay.tv_sec = 0;
-  delay.tv_usec = 0 * 100 * 1000;
-
-  finished = select (mixer_fd + 1, &rfds, NULL, NULL, mixer_needs_resize ? &delay : NULL) < 0;
+  if ((count = snd_mixer_poll_descriptors_count(mixer_handle)) < 0)
+    mixer_abort (ERR_FCN, "snd_mixer_poll_descriptors_count", count);
+  fds = calloc(count + 1, sizeof(struct pollfd));
+  if (fds == NULL)
+    mixer_abort (ERR_FCN, "malloc", 0);
+  fds->fd = fileno(stdin);
+  fds->events = POLLIN;
+  if ((err = snd_mixer_poll_descriptors(mixer_handle, fds + 1, count)) < 0)
+    mixer_abort (ERR_FCN, "snd_mixer_poll_descriptors", err);
+  if (err != count)
+    mixer_abort (ERR_FCN, "snd_mixer_poll_descriptors (err != count)", 0);
+  
+  finished = poll(fds, count + 1, -1);
 
   /* don't abort on handled signals */
-  if (finished && errno == EINTR)
-    {
-      FD_ZERO (&rfds);
-      finished = 0;
-    }
-  else if (mixer_needs_resize)
+  if (finished < 0 && errno == EINTR)
+    finished = 0;
+  if (mixer_needs_resize)
     mixer_resize ();
 
-  if (FD_ISSET (mixer_fd, &rfds))
-    snd_mixer_read (mixer_handle, &callbacks);
+  if (finished > 0) {
+    if (fds->revents & POLLIN) {
+      key = getch ();
+      finished--;
+    }
+  } else {
+    key = 0;
+  }
+  
+  if (finished > 0) {
+    if (snd_mixer_poll_descriptors_revents(mixer_handle, fds + 1, count, &revents) >= 0) {
+      if (revents & POLLNVAL)
+        mixer_abort (ERR_FCN, "snd_mixer_poll_descriptors (POLLNVAL)", 0);
+      if (revents & POLLERR)
+        mixer_abort (ERR_FCN, "snd_mixer_poll_descriptors (POLLERR)", 0);
+      if (revents & POLLIN)
+        snd_mixer_handle_events(mixer_handle);
+    }
+  }
 
-  if (FD_ISSET (fileno (stdin), &rfds))
-    key = getch ();
+  finished = 0;
+  free(fds);
 
   old_view = mixer_view;
   
+#if 0 /* DISABLED: it's not so usefull rather annoying... */
   /* feature Escape prefixing for some keys */
   if (key == 27)
     {
@@ -1460,6 +1734,7 @@ mixer_iteration (void)
 	  break;
 	}
     }
+#endif /* DISABLED */
   
   /* general keys */
   switch (key)
@@ -1489,6 +1764,7 @@ mixer_iteration (void)
       mixer_view = VIEW_PROCINFO;
       key = 0;
       break;
+    case '\014':
     case 'L':
     case 'l':
       mixer_clear (TRUE);
@@ -1675,7 +1951,7 @@ static void
 mixer_signal_handler (int signal)
 {
   if (signal != SIGSEGV)
-    mixer_abort (ERR_SIGNAL, sys_siglist[signal]);
+    mixer_abort (ERR_SIGNAL, sys_siglist[signal], 0);
   else
     {
       fprintf (stderr, "\nSegmentation fault.\n");
@@ -1693,23 +1969,31 @@ main (int    argc,
    */
   do
     {
-      opt = getopt (argc, argv, "c:m:shg");
+      opt = getopt (argc, argv, "c:D:shg");
       switch (opt)
 	{
 	case '?':
 	case 'h':
 	  fprintf (stderr, "%s %s\n", PRGNAME_UPPER, VERSION);
-	  fprintf (stderr, "Usage: %s [-c <card: 1..%i>] [-m <mixer: 0..1>] [-z]\n", PRGNAME, snd_cards ());
-	  mixer_abort (ERR_NONE, "");
+	  fprintf (stderr, "Usage: %s [-h] [-c <card: 0...7 or id>] [-D <mixer device>] [-g] [-s]\n", PRGNAME);
+	  mixer_abort (ERR_NONE, "", 0);
 	case 'c':
-	  card_id = snd_card_name (optarg);
+	  {
+	    int i = snd_card_get_index(optarg);
+	    if (i < 0 || i > 31) {
+	      fprintf (stderr, "wrong -c argument '%s'\n", optarg);
+	      mixer_abort (ERR_NONE, "", 0);
+	    }
+	    sprintf(card_id, "hw:%i", i);
+	  }
+ 	  break;
+	case 'D':
+	  strncpy(card_id, optarg, sizeof(card_id));
+	  card_id[sizeof(card_id)-1] = '\0';
 	  break;
 	case 'g':
 	  mixer_do_color = !mixer_do_color;
 	  break;
-	case 'm':
-	  mixer_id = CLAMP (optarg[0], '0', '1') - '0';
-	  break;
 	case 's':
 	  mixer_minimize = 1;
 	  break;
@@ -1722,11 +2006,16 @@ main (int    argc,
   mixer_init ();
   mixer_reinit ();
   
+  if (mixer_n_elems == 0) {
+    fprintf(stderr, "No mixer elems found\n");
+    mixer_abort (ERR_NONE, "", 0);
+  }
+
   /* setup signal handlers
    */
   signal (SIGINT, mixer_signal_handler);
   signal (SIGTRAP, mixer_signal_handler);
-  signal (SIGABRT, mixer_signal_handler);
+  // signal (SIGABRT, mixer_signal_handler);
   signal (SIGQUIT, mixer_signal_handler);
   signal (SIGBUS, mixer_signal_handler);
   signal (SIGSEGV, mixer_signal_handler);
@@ -1765,5 +2054,5 @@ main (int    argc,
     }
   while (!mixer_iteration ());
   
-  mixer_abort (ERR_NONE, "");
+  mixer_abort (ERR_NONE, "", 0);
 };
